@@ -1,4 +1,4 @@
-from typing import TYPE_CHECKING, List, Type, Union
+from typing import TYPE_CHECKING, List, Sequence, Type, Union
 
 from jinja2 import ChoiceLoader, FileSystemLoader, PackageLoader
 from sqlalchemy.engine import Engine
@@ -6,6 +6,7 @@ from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession
 from sqlalchemy.orm import Session, sessionmaker
 from starlette.applications import Starlette
 from starlette.exceptions import HTTPException
+from starlette.middleware import Middleware
 from starlette.requests import Request
 from starlette.responses import RedirectResponse, Response
 from starlette.routing import Mount, Route
@@ -52,6 +53,7 @@ class BaseAdmin:
         self.templates.env.globals["admin_title"] = title
         self.templates.env.globals["admin_logo_url"] = logo_url
         self.templates.env.globals["model_admins"] = self.model_admins
+        self.templates.env.globals["is_list"] = lambda x: isinstance(x, list)
 
     @property
     def model_admins(self) -> List["ModelAdmin"]:
@@ -89,6 +91,8 @@ class BaseAdmin:
 
         # Set database engine from Admin instance
         model.engine = self.engine
+        model.url_path_for = self.app.url_path_for
+
         if isinstance(model.engine, Engine):
             model.sessionmaker = sessionmaker(bind=model.engine, class_=Session)
             model.async_engine = False
@@ -125,6 +129,13 @@ class BaseAdminView(BaseAdmin):
         if not model_admin.can_edit or not model_admin.is_accessible(request):
             raise HTTPException(status_code=403)
 
+    async def _export(self, request: Request) -> None:
+        model_admin = self._find_model_admin(request.path_params["identity"])
+        if not model_admin.can_export or not model_admin.is_accessible(request):
+            raise HTTPException(status_code=403)
+        if request.path_params["export_type"] not in model_admin.export_types:
+            raise HTTPException(status_code=404)
+
 
 class Admin(BaseAdminView):
     """Main entrypoint to admin interface.
@@ -156,6 +167,8 @@ class Admin(BaseAdminView):
         base_url: str = "/admin",
         title: str = "Admin",
         logo_url: str = None,
+        middlewares: Sequence[Middleware] = None,
+        debug: bool = False,
     ) -> None:
         """
         Args:
@@ -210,8 +223,16 @@ class Admin(BaseAdminView):
                     name="edit",
                     methods=["GET", "POST"],
                 ),
+                Route(
+                    "/{identity}/export/{export_type}",
+                    endpoint=self.export,
+                    name="export",
+                    methods=["GET"],
+                ),
             ],
             exception_handlers={HTTPException: http_exception},
+            middleware=middlewares,
+            debug=debug,
         )
         self.app.mount(base_url, app=admin, name="admin")
 
@@ -352,3 +373,15 @@ class Admin(BaseAdminView):
             request.url_for("admin:list", identity=identity),
             status_code=302,
         )
+
+    async def export(self, request: Request) -> Response:
+        """Export model endpoint."""
+
+        await self._export(request)
+
+        identity = request.path_params["identity"]
+        export_type = request.path_params["export_type"]
+
+        model_admin = self._find_model_admin(identity)
+        rows = await model_admin.get_model_objects(limit=model_admin.export_max_rows)
+        return model_admin.export_data(rows, export_type=export_type)
