@@ -1,9 +1,11 @@
+import time
 from enum import Enum
 from typing import (
     Any,
     Callable,
     ClassVar,
     Dict,
+    Generator,
     List,
     Optional,
     Sequence,
@@ -27,11 +29,19 @@ from sqlalchemy.orm import (
 from sqlalchemy.orm.attributes import InstrumentedAttribute
 from sqlalchemy.sql.elements import ClauseElement
 from starlette.requests import Request
+from starlette.responses import StreamingResponse
 from wtforms import Field, Form
 
 from sqladmin.exceptions import InvalidColumnError, InvalidModelError
 from sqladmin.forms import get_model_form
-from sqladmin.helpers import prettify_class_name, slugify_class_name
+from sqladmin.helpers import (
+    Writer,
+    as_str,
+    prettify_class_name,
+    secure_filename,
+    slugify_class_name,
+    stream_to_csv,
+)
 from sqladmin.pagination import Pagination
 
 __all__ = [
@@ -79,6 +89,9 @@ class ModelAdminMeta(type):
         mcls._check_conflicting_options(
             ["column_details_list", "column_details_exclude_list"], attrs
         )
+        mcls._check_conflicting_options(
+            ["column_export_list", "column_export_exclude_list"], attrs
+        )
 
         return cls
 
@@ -128,6 +141,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
     sessionmaker: ClassVar[sessionmaker]
     engine: ClassVar[Union[Engine, AsyncEngine]]
     async_engine: ClassVar[bool]
+    url_path_for: ClassVar[Callable]
 
     # Metadata
     name: ClassVar[str] = ""
@@ -166,6 +180,11 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
     Default value is set to `True`.
     """
 
+    can_export: ClassVar[bool] = True
+    """Permission for exporting lists of Models.
+    Default value is set to `True`.
+    """
+
     # List page
     column_list: ClassVar[Sequence[Union[str, InstrumentedAttribute]]] = []
     """List of columns to display in `List` page.
@@ -192,6 +211,28 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         ```
     """
 
+    column_formatters: ClassVar[
+        Dict[Union[str, InstrumentedAttribute], Callable[[type, Column], Any]]
+    ] = {}
+    """Dictionary of list view column formatters.
+    Columns can either be string names or SQLAlchemy columns.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelAdmin, model=User):
+            column_formatters = {User.name: lambda m, a: m.name[:10]}
+        ```
+
+    The format function has the prototype:
+    ???+ formatter
+        ```python
+        def formatter(model, attribute):
+            # `model` is model instance
+            # `attribute` is a Union[Column, ColumnProperty, RelationshipProperty]
+            pass
+        ```
+    """
+
     page_size: ClassVar[int] = 10
     """Default number of items to display in `List` page pagination.
     Default value is set to `10`.
@@ -264,6 +305,28 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         ```
     """
 
+    column_formatters_detail: ClassVar[
+        Dict[Union[str, InstrumentedAttribute], Callable[[type, Column], Any]]
+    ] = {}
+    """Dictionary of details view column formatters.
+    Columns can either be string names or SQLAlchemy columns.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelAdmin, model=User):
+            column_formatters_detail = {User.name: lambda m, a: m.name[:10]}
+        ```
+
+    The format function has the prototype:
+    ???+ formatter
+        ```python
+        def formatter(model, attribute):
+            # `model` is model instance
+            # `attribute` is a Union[Column, ColumnProperty, RelationshipProperty]
+            pass
+        ```
+    """
+
     column_labels: ClassVar[Dict[Union[str, InstrumentedAttribute], str]] = {}
     """A mapping of column labels, used to map column names to new names.
     Dictionary keys can be string names or SQLAlchemy columns with string values.
@@ -288,6 +351,39 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
     edit_template: ClassVar[str] = "edit.html"
     """Edit view template. Default is `edit.html`."""
 
+    # Export
+    column_export_list: ClassVar[List[Union[str, InstrumentedAttribute]]] = []
+    """List of columns to include when exporting.
+    Columns can either be string names or SQLAlchemy columns.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelAdmin, model=User):
+            column_export_list = [User.id, User.name]
+        ```
+    """
+
+    column_export_exclude_list: ClassVar[List[Union[str, InstrumentedAttribute]]] = []
+    """List of columns to exclude when exporting.
+    Columns can either be string names or SQLAlchemy columns.
+
+    ???+ example
+        ```python
+        class UserAdmin(ModelAdmin, model=User):
+            column_export_exclude_list = [User.id, User.name]
+        ```
+    """
+
+    export_types: ClassVar[List[str]] = ["csv"]
+    """A list of available export filetypes.
+    Currently only `csv` is supported.
+    """
+
+    export_max_rows: ClassVar[int] = 0
+    """Maximum number of rows allowed for export.
+    Unlimited by default.
+    """
+
     # Form
     form: ClassVar[Optional[Type[Form]]] = None
     """Form class.
@@ -372,6 +468,9 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
     def __init__(self) -> None:
         self._column_labels = self.get_column_labels()
+        self._column_labels_value_by_key = {
+            v: k for k, v in self._column_labels.items()
+        }
 
         self._list_attrs = self.get_list_columns()
         self._list_columns = [
@@ -397,6 +496,22 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             if isinstance(attr, RelationshipProperty)
         ]
 
+        column_formatters = getattr(self, "column_formatters", {})
+        self._list_formatters = {
+            self.get_model_attr(attr): formatter
+            for (attr, formatter) in column_formatters.items()
+        }
+
+        column_formatters_detail = getattr(self, "column_formatters_detail", {})
+        self._detail_formatters = {
+            self.get_model_attr(attr): formatter
+            for (attr, formatter) in column_formatters_detail.items()
+        }
+
+        self._form_attrs = self.get_form_columns()
+
+        self._export_attrs = self.get_export_columns()
+
         self._search_fields = [
             getattr(self.model, self.get_model_attr(attr).key)
             for attr in self.column_searchable_list or []
@@ -448,13 +563,51 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         except NotImplementedError:
             return str
 
+    def _url_for_details(self, obj: Any) -> str:
+        pk = getattr(obj, inspect(obj).mapper.primary_key[0].name)
+        return self.url_path_for(
+            "admin:details",
+            identity=slugify_class_name(obj.__class__.__name__),
+            pk=pk,
+        )
+
+    def _url_for_edit(self, obj: Any) -> str:
+        pk = getattr(obj, inspect(obj).mapper.primary_key[0].name)
+        return self.url_path_for(
+            "admin:edit",
+            identity=slugify_class_name(obj.__class__.__name__),
+            pk=pk,
+        )
+
+    def _url_for_delete(self, obj: Any) -> str:
+        pk = getattr(obj, inspect(obj).mapper.primary_key[0].name)
+        return self.url_path_for(
+            "admin:delete",
+            identity=slugify_class_name(obj.__class__.__name__),
+            pk=pk,
+        )
+
+    def _url_for_details_with_attr(self, obj: Any, attr: RelationshipProperty) -> str:
+        target = getattr(obj, attr.key)
+        pk = getattr(target, attr.mapper.primary_key[0].name)
+        return self.url_path_for(
+            "admin:details",
+            identity=slugify_class_name(target.__class__.__name__),
+            pk=pk,
+        )
+
     async def count(self) -> int:
         stmt = select(func.count(self.pk_column))
         rows = await self._run_query(stmt)
         return rows[0]
 
     async def list(
-        self, page: int, page_size: int, search: str, sort_by: str, sort: str
+        self,
+        page: int,
+        page_size: int,
+        search: Optional[str] = None,
+        sort_by: Optional[str] = None,
+        sort: str = "asc",
     ) -> Pagination:
         page_size = min(page_size or self.page_size, max(self.page_size_options))
 
@@ -484,6 +637,17 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
 
         return pagination
 
+    async def get_model_objects(self, limit: int = 0) -> List[Any]:
+        # For unlimited rows this should pass None
+        limit = None if limit == 0 else limit
+        stmt = select(self.model).limit(limit=limit)
+
+        for _, relation in self._list_relations:
+            stmt = stmt.options(selectinload(relation.key))
+
+        rows = await self._run_query(stmt)
+        return rows
+
     async def get_model_by_pk(self, value: Any) -> Any:
         stmt = select(self.model).where(
             self.pk_column == self._get_column_python_type(self.pk_column)(value)
@@ -502,13 +666,30 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
     ) -> Any:
         if isinstance(attr, Column):
             return getattr(obj, attr.name)
-        else:
-            value = getattr(obj, attr.key)
-            if isinstance(value, list):
-                return ", ".join(map(str, value))
-            elif isinstance(value, Enum):
-                return value.value
-            return value
+
+        value = getattr(obj, attr.key)
+        if isinstance(value, Enum):
+            return value.value
+
+        return value
+
+    def get_list_value(
+        self, obj: type, attr: Union[Column, ColumnProperty, RelationshipProperty]
+    ) -> Any:
+        """Get instancee values for the list view."""
+        formatter = self._list_formatters.get(attr)
+        if formatter:
+            return formatter(obj, attr)
+        return self.get_attr_value(obj, attr)
+
+    def get_detail_value(
+        self, obj: type, attr: Union[Column, ColumnProperty, RelationshipProperty]
+    ) -> Any:
+        """Get instancee values for the detail view."""
+        formatter = self._detail_formatters.get(attr)
+        if formatter:
+            return formatter(obj, attr)
+        return self.get_attr_value(obj, attr)
 
     def get_model_attr(
         self, attr: Union[str, InstrumentedAttribute]
@@ -522,12 +703,16 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         elif isinstance(attr.prop, RelationshipProperty):
             key = attr.prop.key
 
-        try:
+        if key in inspect(self.model).attrs:
             return inspect(self.model).attrs[key]
-        except KeyError:
-            raise InvalidColumnError(
-                f"Model '{self.model.__name__}' has no attribute '{key}'."
-            )
+
+        # Get value by column label
+        if key in self._column_labels_value_by_key:
+            return self._column_labels_value_by_key[key]
+
+        raise InvalidColumnError(
+            f"Model '{self.model.__name__}' has no attribute '{key}'."
+        )
 
     def get_model_attributes(self) -> List[Column]:
         return list(inspect(self.model).attrs)
@@ -588,6 +773,20 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             default=self.get_model_attributes,
         )
 
+    def get_export_columns(self) -> List[Tuple[str, Column]]:
+        """Get list of columns to export."""
+
+        columns = getattr(self, "column_export_list", None)
+        excluded_columns = getattr(self, "column_export_exclude_list", None)
+        if not columns and not excluded_columns:
+            return self.get_list_columns()
+
+        return self._build_column_list(
+            include=columns,
+            exclude=excluded_columns,
+            default=lambda: self._list_columns,
+        )
+
     def get_column_labels(self) -> Dict[Column, str]:
         return {
             self.get_model_attr(column_label): value
@@ -635,7 +834,7 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
         return await get_model_form(
             model=self.model,
             engine=self.engine,
-            only=[i[1].key for i in self.get_form_columns()],
+            only=[i[1].key for i in self._form_attrs],
             column_labels={k.key: v for k, v in self._column_labels.items()},
             form_args=self.form_args,
             form_class=self.form_base_class,
@@ -662,3 +861,43 @@ class ModelAdmin(BaseModelAdmin, metaclass=ModelAdminMeta):
             self._column_labels.get(field, field.key) for field in search_fields
         ]
         return ", ".join(field_names)
+
+    def get_export_name(self, export_type: str) -> str:
+        """The file name when exporting."""
+        filename = f"{self.name}_{time.strftime('%Y-%m-%d_%H-%M-%S')}.{export_type}"
+        return filename
+
+    def export_data(
+        self,
+        data: List[Any],
+        export_type: str = "csv",
+    ) -> StreamingResponse:
+        if export_type == "csv":
+            return self._export_csv(data)
+        else:
+            raise NotImplementedError("Only export_type='csv' is implemented.")
+
+    def _export_csv(
+        self,
+        data: List[Any],
+    ) -> StreamingResponse:
+        def generate(writer: Writer) -> Generator[List[str], None, None]:
+            # Append the column titles at the beginning
+            titles = [c[0] for c in self._export_attrs]
+            yield writer.writerow(titles)
+
+            for row in data:
+                vals = [
+                    as_str(self.get_attr_value(row, c[1])) for c in self._export_attrs
+                ]
+                yield writer.writerow(vals)
+
+        # `get_export_name` can be subclassed.
+        # So we want to keep the filename secure outside that method.
+        filename = secure_filename(self.get_export_name(export_type="csv"))
+
+        return StreamingResponse(
+            content=stream_to_csv(generate),
+            media_type="text/csv",
+            headers={"Content-Disposition": f"attachment;filename={filename}"},
+        )
