@@ -25,7 +25,7 @@ $(document).on('click','#search-button',function() {
 
   newUrl = "";
   if (window.location.search && window.location.search.indexOf('search=') != -1) {
-    newUrl = window.location.search.replace( /search=\w*\d*/, "search=" + searchTerm);
+    newUrl = window.location.search.replace( /search=[^&]*/, "search=" + searchTerm);
   } else if (window.location.search) {
     newUrl = window.location.search + "&search=" + searchTerm;
   } else {
@@ -37,6 +37,13 @@ $(document).on('click','#search-button',function() {
 // Reset search
 $(document).on('click','#search-reset',function() {
   if (window.location.search && window.location.search.indexOf('search=') != -1) {
-    window.location.href = window.location.search.replace( /search=\w*\d*/, "");
+    window.location.href = window.location.search.replace( /search=[^&]*/, "");
+  }
+});
+
+// Press enter to search
+$(document).on('keypress','#search-input',function(e) {
+  if (e.which === 13) {
+    $('#search-button').click();
   }
 });
